from joblib import Parallel as _Parallel
from joblib import delayed as _delayed
from joblib import cpu_count as _cpu_count
from time import perf_counter as _perf_counter
import numpy as _np
import ctypes as _ctypes
import os as _os
from pathlib import Path as _Path

_current_dir = _Path(_os.path.dirname(_os.path.realpath(__file__)))
_cudalib_path = _current_dir / "cuda-libraries" / "cudalib.dll"
cudalib = _ctypes.cdll.LoadLibrary(str(_cudalib_path))

_current_dir = _Path(_os.path.dirname(_os.path.realpath(__file__)))
_clib_path = _current_dir / "c-libraries" / "clib.dll"
clib = _ctypes.cdll.LoadLibrary(str(_clib_path))

def map(func, iterable, num_cores:int=None, log_time:bool=False):
    """
    Applies a function to all items in an input iterable using multiple cores.

    This function is a parallelized version of the built-in map() function, 
    which applies a function to all items in an iterable. The function is 
    applied in parallel using multiple cores for increased performance.

    Parameters:
        func (callable): The function to apply to the items.
        iterable (iterable): An iterable of items to process.
        num_cores (int, optional): The number of cores to use. If None, all available cores are used. 
        If the number is greater than the number of available cores, a ValueError is raised.
        log_time (bool, optional): If True, the time taken to apply the function is printed.

    Returns:
        iterable: An iterable of the same type as the input, containing the results of applying the function.

    Raises:
        ValueError: If num_cores is greater than the number of available cores.
    """
    if num_cores is None:
        num_cores = _cpu_count()
    elif num_cores > _cpu_count():
        raise ValueError (f"Number of cores cannot be greater than the number of cores on your system: {_cpu_count()}")
    
    start_time = _perf_counter()

    return_type = type(iterable)
    if num_cores == 1:
        res = return_type(__builtins__['map'](func, iterable))
    else:
        res = return_type(_Parallel(n_jobs=num_cores)(_delayed(func)(i) for i in iterable))

    end_time=_perf_counter()
    if log_time:
        print(f"Time Elapsed for map(): {end_time-start_time} seconds.")
    return res

def apply_map(func, iterable, num_cores:int=None, log_time:bool=False) -> None:
    """
    Applies a function to all items in an input iterable using multiple cores and modifies the iterable in-place.

    This function is a parallelized version of the built-in map() function, 
    which applies a function to all items in an iterable. The function is 
    applied in parallel using multiple cores for increased performance. 
    The results are used to modify the input iterable in-place.

    Parameters:
        func (callable): The function to apply to the items.
        iterable (iterable): An iterable of items to process. This iterable is modified in-place.
        num_cores (int, optional): The number of cores to use. If None, all available cores are used. 
        If the number is greater than the number of available cores, a ValueError is raised.
        log_time (bool, optional): If True, the time taken to apply the function is printed.

    Returns:
        

    Raises:
        ValueError: If num_cores is greater than the number of available cores.
    """
    if num_cores is None:
        num_cores = _cpu_count()
    elif num_cores > _cpu_count():
        raise ValueError (f"Number of cores cannot be greater than the number of cores on your system: {_cpu_count()}")
    
    start_time = _perf_counter()

    if num_cores == 1:
        res = list(__builtins__['map'](func, iterable))
    else:
        res = list(_Parallel(n_jobs=num_cores)(_delayed(func)(i) for i in iterable))

    for i in range (iterable):
        iterable[i] = res[i]

    end_time=_perf_counter()
    if log_time:
        print(f"Time Elapsed for map(): {end_time-start_time} seconds.")
    

def filter(func_bool, iterable, num_cores:int=None, log_time:bool=False):
    """
    Filters items in an input iterable using a boolean function and multiple cores.

    This function is a parallelized version of the built-in filter() function, 
    which filters items in an iterable based on a boolean function. The function 
    is applied in parallel using multiple cores for increased performance. 

    Parameters:
        func_bool (callable): The boolean function to apply to the items. This function should return a boolean value.
        iterable (iterable): An iterable of items to process.
        num_cores (int, optional): The number of cores to use. If None, all available cores are used. 
        If the number is greater than the number of available cores, a ValueError is raised.
        log_time (bool, optional): If True, the time taken to apply the function is printed.

    Returns:
        return_type: A new iterable of the same type as the input, containing only the items for which the boolean function returned True.

    Raises:
        ValueError: If num_cores is greater than the number of available cores.
        ValueError: If func_bool does not return a boolean value.
    """

    if num_cores is None:
        num_cores = _cpu_count()
    elif num_cores > _cpu_count():
        raise ValueError (f"Number of cores cannot be greater than the number of cores on your system: {_cpu_count()}")
    
    bool_check = func_bool(iterable[0])
    if not isinstance(bool_check, bool):
        raise ValueError("Function argument 'func_bool' must return a boolean value")
    
    start_time = _perf_counter() 

    return_type = type(iterable)
    if num_cores == 1:
        res = return_type(__builtins__['filter'](func_bool, iterable))
    else:
        res = _Parallel(n_jobs=num_cores)(_delayed(lambda x: (x, func_bool(x)))(i) for i in iterable)
        res = return_type([x for x, keep in res if keep])
    end_time = _perf_counter()
    if log_time:
        print(f"Time Elapsed for filter(): {end_time-start_time} seconds.")

    return res

def frequency(iterable, log_time=False) -> dict:
    """
    Computes the frequency of items in an iterable.

    This function iterates over the input iterable and computes the frequency of each item. 
    The frequencies are returned as a dictionary where the keys are the items and the values are the frequencies.

    Parameters:
        iterable (iterable): An iterable of items to compute frequencies for.
        log_time (bool, optional): If True, the time taken to compute the frequencies is printed.

    Returns:
        dict: A dictionary where the keys are the items from the iterable and the values are the frequencies.
    """
    start_time = _perf_counter()
    freq = {}
    for i in iterable:
        if i not in freq:
            freq[i] = 1
        else:
            freq[i] += 1
    end_time = _perf_counter()

    if log_time:
        print(f"Time Elapsed for get_freq(): {end_time-start_time} seconds.")

    return freq

def sum(iterable, num_cores=None, log_time=False) -> float:
    """
    Computes the mean of an iterable.

    This function uses multiple cores to compute the mean of an iterable.

    Parameters:
        iterable (iterable): An iterable of numbers to compute the mean of.
        num_cores (int, optional): The number of cores to use. If None, 
        len(iterable) will determine number of cores are used.
        log_time (bool, optional): If True, the time taken to compute the mean is printed.

    Returns:
        float: The mean of the iterable.
    """
    if num_cores is None:
        num_cores = _cpu_count()
    elif num_cores > _cpu_count():
        raise ValueError (f"Number of cores cannot be greater than the number of cores on your system: {_cpu_count()}")
    
    start_time = _perf_counter()
    if not isinstance(iterable, list):
        try:
            iterable = _np.array(iterable)
        except Exception:
            pass
    if isinstance(iterable, _np.ndarray):
        return _np.sum(iterable)
    else:
        if len(iterable) < 500_000_000 or num_cores==1:
            res = __builtins__['sum'](iterable)
        else:
            lst = [iterable[i[0]:i[1]] for i in __partition(len(iterable), num_cores)]
            lst = map(__builtins__['sum'], lst, num_cores)
            res = __builtins__['sum'](lst)

    end_time = _perf_counter()
    if log_time:
        print(f"Time Elapsed for sum(): {end_time-start_time} seconds.")
    
    return res

def mean(iterable, num_cores=None, log_time=False) -> float:
    """
    Computes the mean of an iterable.

    This function uses multiple cores to compute the mean of an iterable.

    Parameters:
        iterable (iterable): An iterable of numbers to compute the mean of.
        num_cores (int, optional): The number of cores to use. If None, len(iterable) 
        will determine number of cores are used.
        log_time (bool, optional): If True, the time taken to compute the mean is printed.

    Returns:
        float: The mean of the iterable.
    """
    start_time = _perf_counter()
    try:
        iterable = _np.array(iterable)
    except Exception:
        pass

    if isinstance(iterable, _np.ndarray):
        res =  _np.mean(iterable)
    else:
        total = sum(iterable=iterable, num_cores=num_cores)
        res = total / len(iterable)
    end_time = _perf_counter()

    if log_time:
        print(f"Time Elapsed for avg(): {end_time-start_time} seconds.")

    return res

def var(iterable, num_cores=None, log_time=False):
    """
    Computes the svarience of an iterable.

    This function uses multiple cores to compute the varience of an iterable.

    Parameters:
        iterable (iterable): An iterable of numbers to compute the varience of.
        num_cores (int, optional): The number of cores to use. If None, all available cores are used.
        log_time (bool, optional): If True, the time taken to compute the varience is printed.

    Returns:
        float: The varience of the iterable.
    """
    if num_cores is None:
        if len(iterable) < 500_000:
            num_cores = 1
        else:
            num_cores = _cpu_count()
    elif num_cores > _cpu_count():
        raise ValueError (f"Number of cores cannot be greater than the number of cores on your system: {_cpu_count()}")
    
    def sum_squared_errors(x):
        diff = x - avg
        return (diff**2)

    start_time = _perf_counter()

    try:
        iterable = _np.array(iterable)
    except Exception:
        pass

    if type(iterable) == _np.ndarray:
        arr_ptr = iterable.astype(_np.double).ctypes.data_as(_ctypes.POINTER(_ctypes.c_double))
        res = cudalib.varience(arr_ptr, len(iterable), 'a')
    elif num_cores == 1:
        avg = mean(lst)
        lst = [(i-avg)**2 for i in lst]
        return sum(lst) / len(lst)
    else:
        avg = mean(iterable, num_cores)
        sum_squared = _Parallel(n_jobs=num_cores)(_delayed(sum_squared_errors)(i) for i in iterable)
        res = mean(sum_squared, num_cores)

    end_time = _perf_counter()
    if log_time:
        print(f"Time Elapsed for sum(): {end_time-start_time} seconds.")
    
    return res


def std(iterable, num_cores:int=None, log_time:bool=False) -> float:
    """
    Computes the standard deviation of an iterable.

    This function uses multiple cores to compute the standard deviation of an iterable.

    Parameters:
        iterable (iterable): An iterable of numbers to compute the standard deviation of.
        num_cores (int, optional): The number of cores to use. If None, all available cores are used.
        log_time (bool, optional): If True, the time taken to compute the standard deviation is printed.

    Returns:
        float: The standard deviation of the iterable.
    """
    start_time = _perf_counter()
    res = var(iterable, num_cores)
    res = res**0.5

    end_time = _perf_counter()
    if log_time:
        print(f"Time Elapsed for sum(): {end_time-start_time} seconds.")
    
    return res


def cpu_count(only_physical_cores=False):
    """
    Returns the number of CPU cores available.

    This function returns the number of CPU cores available on the system. 
    If only_physical_cores is True, it returns the number of physical cores only.

    Parameters:
        only_physical_cores (bool, optional): If True, only physical cores are counted.

    Returns:
        int: The number of CPU cores available.
    """
    return _cpu_count(only_physical_cores=only_physical_cores)

def measure_time(func, args=[], num_trials:int=1):
    """
    Measures and returns the average execution time of a function over a given number of trials.

    This function measures the execution time of the passed function using `time.perf_counter()` method, 
    which provides the highest available resolution timer in the system. 
    The function is executed for a specified number of trials, and the average execution time is returned.

    Parameters:
        func (callable): The function for which the execution time is to be measured.
        args (iterable, optional): Iterable containing arguments to be passed to `func`. Each item in the iterable 
        is passed as a separate argument to the function. Defaults to an empty list.
        num_trials (int, optional): The number of times the function is to be executed for time measurement. Defaults to 1.

    Returns:
        float: The average execution time of the function across all trials, in seconds.

    Raises:
        TypeError: If the 'args' parameter is not an iterable object.
    """
    try:
        args = list(args)
    except:
        raise TypeError(f"Argument passed to 'args' must be an iterable object. Value Recived: {args}")

    trials = []
    while len(trials) < num_trials:
        start = _perf_counter()
        func(*args)
        end = _perf_counter()
        trials.append(end-start)
    return mean(trials)

def num_gpu():
    return max(cudalib.cuda_enabled(), 0)

def __partition(num, num_cores):
    dif = num // num_cores
    partitions = []
    curr = 0
    for i in range (num_cores):
        if i != num_cores-1:
            partitions.append((curr, curr+dif))
            curr += dif
        else:
            partitions.append((curr, num))
    
    return partitions
